//! Bits
//!
//! Handles if a certain bit is set. Vec<bool>

use std::fmt;
use std::ops::RangeBounds;
// use std::iter::{Map, Filter, Enumerate};

/// Flags - track what is set
///
/// A vector of bool with helper functions to track what is set (true) and not (false).
#[derive(Clone)]
pub struct Flags(pub Vec<bool>);

impl Flags {
    /// Construct tracking 0..size number of items.
    pub fn new(size: usize) -> Self {
        Flags(vec![false; size])
    }

    /// Clear all (reset to false)
    #[inline]
    pub fn clear(&mut self) {
        self.0.fill(false);
        /*
        for i in 0..self.0.len() {
            self.0[i] = false;
        }
        */
    }

    /// Fill all (true or false value)
    #[inline]
    pub fn fill(&mut self, value: bool) {
        self.0.fill(value);
        /*
        for i in 0..self.0.len() {
            self.0[i] = value;
        }
        */
    }

    /// Set a position with state (true/false)
    #[inline]
    pub fn set(&mut self, pos: usize, state: bool) {
        self.0[pos] = state;
    }

    /// Return state of position
    #[inline]
    pub fn get(&self, pos: usize) -> bool {
        self.0[pos]
    }

    /// Set a given range with true
    pub fn set_range<R: RangeBounds<usize> + IntoIterator<Item = usize>>(&mut self, bits: R) {
        for i in bits {
            self.0[i as usize] = true;
        }
    }

    /// Count the number of items with state=true
    pub fn count_set(&self) -> usize {
        self.0.iter().filter(|&x| *x).count()
    }

    // https://depth-first.com/articles/2020/06/22/returning-rust-iterators/

    #[deprecated = "Use iter() instead."]
    /// Returns the indexes of items that are true
    pub fn indexes(&self) -> impl Iterator<Item = usize> + '_ {
        self.0.iter().enumerate().filter(|x| *x.1).map(|x| x.0)
    }

    /// Iterate over all of the indexes that are set true.
    pub fn iter(&self) -> impl Iterator<Item = usize> + '_ {
        self.0.iter().enumerate().filter(|x| *x.1).map(|x| x.0)
    }

    // Nope.  This isn't how the original iterator for GenBits worked.
    /*
    pub fn iter(&self) -> std::slice::Iter<'_, bool> {
        self.0.iter()
    }
    */

    #[must_use]
    /// Display bits that are set.
    /// +1 is added to the display, so it matches the cell values (1..N)
    pub fn display(&self) -> String {
        self.0
            .iter()
            .enumerate()
            .filter(|x| *x.1)
            .map(|x| x.0)
            .map(|i| (i + 1).to_string())
            // .map(|i| i.to_string())
            .collect::<Vec<_>>()
            .join(",")
    }
}

impl fmt::Debug for Flags {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(
            f,
            "({})",
            self.0
                .iter()
                .enumerate()
                .filter(|x| *x.1)
                .map(|x| x.0.to_string())
                .collect::<Vec<_>>()
                .join(",")
        )
    }
}

impl fmt::Display for Flags {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(
            f,
            "({})",
            self.0
                .iter()
                .enumerate()
                .filter(|x| *x.1)
                .map(|x| x.0.to_string())
                .collect::<Vec<_>>()
                .join(",")
        )
    }
}

impl PartialEq for Flags {
    fn eq(&self, other: &Self) -> bool {
        if self.0.len() != other.0.len() {
            return false;
        }
        for i in 0..self.0.len() {
            if self.0[i] != other.0[i] {
                return false;
            }
        }
        true
    }
}

// Experiment: Add accessing by []

use std::ops::{Index, IndexMut};

impl Index<usize> for Flags {
    type Output = bool;

    fn index<'a>(&'a self, i: usize) -> &'a bool {
        &self.0[i]
    }
}

impl IndexMut<usize> for Flags {
    fn index_mut<'a>(&'a mut self, i: usize) -> &'a mut bool {
        &mut self.0[i]
    }
}

#[cfg(test)]
mod tests {
    use crate::flags::*;

    #[test]
    fn check_u16() {
        const SIZE: usize = 9;
        let mut p = Flags::new(SIZE);

        p.clear();

        for i in 0..9 {
            let mut result = p.get(i);
            assert_eq!(result, false);
            p.set(i, true);
            result = p.get(i);
            assert_eq!(result, true);
        }

        p.clear();

        // Check access by []
        for i in 0..9 {
            let mut result = p[i];
            assert_eq!(result, false);
            p[i] = true;
            result = p[i];
            assert_eq!(result, true);
        }

        p = Flags::new(SIZE);
        p.set(3, true);
        p.set(5, true);
        p.set(6, true);
        assert_eq!(3, p.count_set());
        let values: Vec<usize> = p.iter().collect();
        assert_eq!(values, vec!(3, 5, 6));
        assert_eq!(3, p.count_set());
        p.set(0, true);
        assert_eq!(4, p.count_set());

        p = Flags::new(SIZE);
        p.set_range(0..6);

        for i in 0..6 {
            let result = p.get(i);
            assert_eq!(result, true);
        }
        assert_eq!(p.get(6), false);
    }

    #[test]
    fn check_u32() {
        const SIZE: usize = 29;
        let mut p = Flags::new(SIZE);

        p.clear();

        for i in 0..29 {
            let mut result = p.get(i);
            assert_eq!(result, false);
            p.set(i, true);
            result = p.get(i);
            assert_eq!(result, true);
        }

        p = Flags::new(SIZE);
        p.set(13, true);
        p.set(15, true);
        p.set(26, true);
        assert_eq!(3, p.count_set());
        let values: Vec<usize> = p.iter().collect();
        assert_eq!(values, vec!(13, 15, 26));
        assert_eq!(3, p.count_set());
        p.set(0, true);
        assert_eq!(4, p.count_set());

        p = Flags::new(SIZE);
        p.set_range(10..26);

        for i in 10..26 {
            let result = p.get(i);
            assert_eq!(result, true);
        }
        assert_eq!(p.get(9), false);
        assert_eq!(p.get(27), false);
    }
}

/*
    // -> Map<Filter<Enumerate<std::slice::Iter<'_, bool>>>>
    pub fn indexes(&self) -> std::slice::Iter<usize>  {
        self.data.iter().enumerate().filter(|x| *x.1).map(|x| x.0)
    }
*/
