//! Ksudoku file support
//!
//! This handles saving and loading ksudoku save files.

// Read a Ksudoku's XML save file.
use std::collections::HashMap;
use std::error;
// use std::fmt;
use std::fs::File;
use std::io::BufReader;
use xml::reader::{EventReader, XmlEvent};
use xml::writer::XmlEvent as WrXmlEvent;
use xml::EmitterConfig;

/*
What I want from load_ksudoku:

  - Puzzle string
  - Solution string (so I can validate it/check my code)
  - specific-type & type (combined as a string is ok - I think)
  - order (as a number)

I get values (the puzzle), I also need solution.

Having order will allow me to use AnyBoard::new(order) and load it up.

"Hints"
These are under the <history> section of the XML.
They consist of simple-event elements containing index and markers.
Index is based on ksudoku's layout (Top-Left to Bottom-Right).
Markers is a binary string of 0's and 1's.  1 indicates that value is possible.
*Note*  The index can be in the history section more then once, we would want
the last markers string for any given index.
*/

/// Ksudoku data structure
///
/// This has most everything contained within the ksudoku save file.
/// I don't extract if they had-help, or msecs-elapsed time.
#[derive(Debug)]
pub struct Ksudoku {
    /// puzzle string:
    /// _ is blank, starts with 'b' (for 1), 'c' for 2.
    pub puzzle: String,
    /// This contains the solution to the puzzle.
    pub solution: String,
    /// "Plain"
    pub specific_type: String,
    /// "sudoku"
    pub puzzle_type: String,
    /// Puzzle width (9, 16, 25).
    pub order: u8,
    /// hints/possible that the user has set in the puzzle.
    pub hints: HashMap<u16, Vec<u16>>,
}

impl Default for Ksudoku {
    fn default() -> Self {
        Ksudoku {
            puzzle: String::new(),
            solution: String::new(),
            specific_type: String::from("Plain"),
            puzzle_type: String::from("sudoku"),
            order: 0,
            hints: HashMap::new(),
        }
    }
}

impl Ksudoku {
    /// Convert (column, row) to (row, column) based index
    ///
    /// This converts from ksudoku style
    fn col_to_row_index(&self, kindex: u16) -> u16 {
        let size = self.order as u16;
        let (y, x) = (kindex % size, kindex / size);
        x + y * size
    }

    /// Convert (row, column) to (column row) based index
    fn row_to_col_index(&self, index: u16) -> u16 {
        let size = self.order as u16;
        let (x, y) = (index % size, index / size);
        y * x * size
    }
}

/// Convert ksudoku index to AnyBoard index value.
///
/// Because ksudoku stores values TL->BR
/// and AnyBoard uses TL->RB, we need to do some math here.
#[deprecated = "Use ksudoku col_to_row_index"]
pub fn ksudoku_to_board(kindex: u16, size: u8) -> u16 {
    let size = size as u16;
    let (y, x) = (kindex % size, kindex / size);
    x + y * size
}

#[must_use]
/// Load ksudoku file, returning Ksudoku structure.
/// - It is up to the caller to determine if they can handle the types
///   and size (order).
pub fn load_ksudoku(filename: std::path::PathBuf) -> Result<Ksudoku, Box<dyn error::Error>> {
    let fh = File::open(filename)?;
    let buffer = BufReader::new(fh);
    let parser = EventReader::new(buffer);
    let mut element_name: String = String::new();
    let mut details = Ksudoku::default();

    for e in parser {
        match e {
            Ok(XmlEvent::StartElement {
                name, attributes, ..
            }) => {
                element_name = name.local_name.clone();
                if element_name == "graph" {
                    // Check the attributes here
                    // <graph specific-type="Plain" type="sudoku" order="9"/>
                    // {"specific-type": "Plain", "type": "sudoku", "order": "9"}

                    let mut attrs: HashMap<String, String> = HashMap::new();

                    for a in &attributes {
                        attrs.insert(a.name.to_string(), a.value.clone());
                    }

                    let blank = String::new();

                    details.specific_type = attrs
                        .get(&"specific-type".to_string())
                        .unwrap_or(&blank)
                        .clone();
                    details.puzzle_type = attrs.get(&"type".to_string()).unwrap_or(&blank).clone();
                    let order = attrs.get(&"order".to_string()).unwrap_or(&blank).clone();

                    if !order.is_empty() {
                        let r = order.parse::<u8>();
                        if r.is_ok() {
                            details.order = r.unwrap();
                        } else {
                            // Failed to convert order to u8...
                            return Err(Box::new(r.unwrap_err()));
                        }
                    }
                }

                // hints the user set in the game.

                if element_name == "simple-event" {
                    let mut index = String::new();
                    let mut markers = String::new();
                    for a in &attributes {
                        let name = a.name.to_string();
                        if name == "index" {
                            index = a.value.clone();
                        }
                        if name == "markers" {
                            markers = a.value.clone();
                        }
                    }

                    if let Ok(idx) = index.parse::<u16>() {
                        let board_index = details.col_to_row_index(idx);
                        // let board_index = ksudoku_to_board(idx, details.order) as u16;
                        if markers.contains('1') {
                            // Yes, string contains a hint/possible.
                            details.hints.insert(
                                board_index,
                                markers
                                    .chars()
                                    .enumerate()
                                    .filter(|(_i, c)| *c == '1')
                                    .map(|(i, _c)| i as u16)
                                    .collect(),
                            );
                        } else {
                            // String of zeros, remove the entry.
                            details.hints.remove(&idx);
                        }
                    }

                    // println!("history: {} : {}", index, markers);
                    // println!("{:?}", details.hints);
                }
            }
            Ok(XmlEvent::Characters(text)) => {
                if element_name == "values" {
                    details.puzzle = text.clone();
                }
                if element_name == "solution" {
                    details.solution = text.clone();
                }
            }
            Err(e) => {
                return Err(Box::new(e));
            }
            // Cover the remaining arms in the match.
            _ => {}
        }
    }

    Ok(details)
}

/// Write out a ksudoku puzzle file
pub fn save_ksudoku(
    filename: std::path::PathBuf,
    data: &Ksudoku,
) -> Result<(), Box<dyn error::Error>> {
    let fh = File::create(filename)?;
    let mut writer = EmitterConfig::new().perform_indent(true).create_writer(fh);

    let mut event: WrXmlEvent = WrXmlEvent::start_element("ksudoku").into();
    writer.write(event)?;
    event = WrXmlEvent::start_element("game")
        .attr("had-help", "0")
        .attr("msecs-elapsed", "0")
        .into();
    writer.write(event)?;
    event = WrXmlEvent::start_element("puzzle").into();
    writer.write(event)?;

    // Convert order to a string.
    let order = data.order.to_string();

    event = WrXmlEvent::start_element("graph")
        .attr("specific-type", &data.specific_type)
        .attr("type", &data.puzzle_type)
        .attr("order", &order)
        .into();
    writer.write(event)?;
    event = WrXmlEvent::end_element().into();
    writer.write(event)?;

    // values (puzzle)
    event = WrXmlEvent::start_element("values").into();
    writer.write(event)?;
    event = WrXmlEvent::characters(&data.puzzle).into();
    writer.write(event)?;
    event = WrXmlEvent::end_element().into();
    writer.write(event)?;

    // solution
    event = WrXmlEvent::start_element("solution").into();
    writer.write(event)?;
    event = WrXmlEvent::characters(&data.solution).into();
    writer.write(event)?;
    event = WrXmlEvent::end_element().into();
    writer.write(event)?;

    // Apparently, the events are consumed...

    event = WrXmlEvent::end_element().into();
    writer.write(event)?;
    event = WrXmlEvent::end_element().into();
    writer.write(event)?;
    event = WrXmlEvent::end_element().into();
    writer.write(event)?;

    Ok(())
}

#[cfg(test)]

mod tests {
    use crate::ksudoku::*;
    use crate::sudoku::*;
    use std::path::PathBuf;

    fn test_files_path() -> PathBuf {
        let mut d = PathBuf::from(env!("CARGO_MANIFEST_DIR"));
        d.push("test_files");
        d
    }

    #[test]
    fn load_hints() {
        let mut testfile = test_files_path();
        testfile.push("hints-puzzle");

        let result = load_ksudoku(testfile);
        assert!(result.is_ok());
        let result = result.unwrap();
        assert!(result.puzzle_type == "sudoku");
        assert!(result.specific_type == "Plain");

        let size = (result.order as f64).sqrt() as u8;
        let mut board = AnyBoard::new(size);
        let load = board.load_ksudoku(&result.puzzle);
        assert!(load.is_ok());

        // Ok, puzzle is loaded!
        let strings = board.to_strings();
        assert_eq!(
            strings,
            vec![
                "7 6   1 9",
                "13  7  54",
                "    1    ",
                "6   9   7",
                " 5 4 7 8 ",
                "  9   5  ",
                "   3 9   ",
                "  71 52  ",
                " 4     6 ",
            ]
        );

        assert_eq!(
            result.hints,
            HashMap::from([
                (12, vec![8]),
                (18, vec![3, 4]),
                (19, vec![8]),
                (20, vec![3, 4]),
                (21, vec![8]),
                (24, vec![6]),
                (25, vec![6]),
                (32, vec![0]),
                (42, vec![8]),
                (45, vec![3]),
                (46, vec![6]),
                (50, vec![0]),
                (52, vec![8]),
                (55, vec![5]),
                (58, vec![3, 5]),
                (64, vec![5]),
                (67, vec![3, 5]),
                (70, vec![8]),
                (72, vec![8]),
                (75, vec![6]),
                (78, vec![8]),
                (80, vec![4]),
            ])
        );
    }
    #[test]
    fn load_puzzles() {
        let mut testfile = test_files_path();
        testfile.push("missing");

        let result = load_ksudoku(testfile);
        assert!(result.is_err());

        let mut testfile = test_files_path();
        testfile.push("9-puzzle");

        let result = load_ksudoku(testfile);
        assert!(result.is_ok());
        let result = result.unwrap();
        assert!(result.puzzle_type == "sudoku");
        assert!(result.specific_type == "Plain");

        let size = (result.order as f64).sqrt() as u8;
        let mut board = AnyBoard::new(size);
        let load = board.load_ksudoku(&result.puzzle);
        assert!(load.is_ok());

        // Ok, puzzle is loaded!
        let strings = board.to_strings();
        assert_eq!(
            strings,
            vec![
                "  4   8  ",
                " 6  1  2 ",
                "3  9 5  6",
                "  72 16  ",
                " 3     9 ",
                "  18 35  ",
                "9  3 6  2",
                " 4  5  8 ",
                "  3   7  ",
            ]
        );

        let mut testfile = test_files_path();
        testfile.push("16-puzzle");

        let result = load_ksudoku(testfile);
        assert!(result.is_ok());
        let result = result.unwrap();
        assert!(result.puzzle_type == "sudoku");
        assert!(result.specific_type == "Plain");

        let size = (result.order as f64).sqrt() as u8;
        let mut board = AnyBoard::new(size);
        let load = board.load_ksudoku(&result.puzzle);
        assert!(load.is_ok());

        // Ok, puzzle is loaded!
        let strings: Vec<String> = board.to_strings();
        assert_eq!(
            strings,
            vec![
                " D    O    C  IN",
                "A  ENC   IL     ",
                "NG  AP   J MHC  ",
                "  P H   D A  FOL",
                "IOHKFB  A   L  P",
                " BN  M E L ID   ",
                "LE  O AN K BC   ",
                " C    H    JO EF",
                "EN GP    A    F ",
                "   OG J IM L  NC",
                "   MK B C N  PG ",
                "C  L   F  PEMIKO",
                "OPC  N H   F J  ",
                "  EHJ I   MA  CK",
                "     FM   IPA  D",
                "FM  L    H    P "
            ]
        );
        // board.display();

        let mut testfile = test_files_path();
        testfile.push("25-puzzle");

        let result = load_ksudoku(testfile);
        assert!(result.is_ok());
        let result = result.unwrap();
        assert!(result.puzzle_type == "sudoku");
        assert!(result.specific_type == "Plain");

        let size = (result.order as f64).sqrt() as u8;
        let mut board = AnyBoard::new(size);
        let load = board.load_ksudoku(&result.puzzle);
        assert!(load.is_ok());

        // Ok, puzzle is loaded!
        // board.display();
        let strings: Vec<String> = board.to_strings();
        assert_eq!(
            strings,
            vec![
                " T DPF    Y H R    WSX L ",
                "L QF     J X C G     UB D",
                " CK S LNRP  G  UTQO H MA ",
                "JG   H S XELDUPM F B   RT",
                "Y N RQ UCDOK AISJL HP G E",
                "F  OA N UK VQI HY B DT  C",
                "  S  A C VUE GJQ N I  R  ",
                "  CPD JGMIA   OELSU VBK  ",
                "  G LE D BHT XMW K PI Y  ",
                " EXIBOWFLY     VAMTJUGQS ",
                "G  HV TMI       EWF BR  O",
                " A JFUK W  P D  M GLXE N ",
                "R LN G             O QI F",
                " S TCYP B  H O  X JNAK M ",
                "M  XK ALJ       UHQ GP  Y",
                " VTRYBSEFM     KWOICJNLG ",
                "  U XD J WCN RTA E QY P  ",
                "  HQN COVTJ   EBGDL WSF  ",
                "  F  X Y LWB SVJ R T  O  ",
                "E  CJ R AN GOF XH V MD  B",
                "S V OI ANHDC TGLQJ YF X P",
                "CX   L K RFUYBWO V A   DQ",
                " FR H DQOC  K  INBW T UY ",
                "T JL     G I P F     OC W",
                " B KMV    Q J H    GRI E "
            ]
        );
    }

    #[test]
    fn solve_puzzles() {
        let mut testfile = test_files_path();
        testfile.push("9-puzzle");

        let result = load_ksudoku(testfile);
        assert!(result.is_ok());
        let result = result.unwrap();
        assert!(result.puzzle_type == "sudoku");
        assert!(result.specific_type == "Plain");

        let size = (result.order as f64).sqrt() as u8;
        let mut board = AnyBoard::new(size);
        let load = board.load_ksudoku(&result.puzzle);
        let mut solution = AnyBoard::new(size);
        let solution_load = solution.load_ksudoku(&result.solution);
        assert!(load.is_ok());
        assert!(solution_load.is_ok());

        // Ok, puzzle is loaded!
        let strings = board.to_strings();
        assert_eq!(
            strings,
            vec![
                "  4   8  ",
                " 6  1  2 ",
                "3  9 5  6",
                "  72 16  ",
                " 3     9 ",
                "  18 35  ",
                "9  3 6  2",
                " 4  5  8 ",
                "  3   7  ",
            ]
        );
        // Verify the solution in the file.
        let strings = solution.to_strings();
        assert_eq!(
            strings,
            vec![
                "714632859",
                "569718324",
                "328945176",
                "457291638",
                "836574291",
                "291863547",
                "975386412",
                "642157983",
                "183429765",
            ]
        );

        // Solve puzzle
        let mut solver = AnySolver::new_from(&board);
        assert!(solver.solve_logic());
        assert_eq!(solver.board, solution);

        let mut testfile = test_files_path();
        testfile.push("16-puzzle");

        let result = load_ksudoku(testfile);
        assert!(result.is_ok());
        let result = result.unwrap();
        assert!(result.puzzle_type == "sudoku");
        assert!(result.specific_type == "Plain");

        let size = (result.order as f64).sqrt() as u8;
        let mut board = AnyBoard::new(size);
        let load = board.load_ksudoku(&result.puzzle);
        assert!(load.is_ok());
        let mut solution = AnyBoard::new(size);
        let solution_load = solution.load_ksudoku(&result.solution);
        assert!(solution_load.is_ok());

        // Ok, puzzle is loaded!
        let strings: Vec<String> = board.to_strings();
        assert_eq!(
            strings,
            vec![
                " D    O    C  IN",
                "A  ENC   IL     ",
                "NG  AP   J MHC  ",
                "  P H   D A  FOL",
                "IOHKFB  A   L  P",
                " BN  M E L ID   ",
                "LE  O AN K BC   ",
                " C    H    JO EF",
                "EN GP    A    F ",
                "   OG J IM L  NC",
                "   MK B C N  PG ",
                "C  L   F  PEMIKO",
                "OPC  N H   F J  ",
                "  EHJ I   MA  CK",
                "     FM   IPA  D",
                "FM  L    H    P "
            ]
        );

        let strings = solution.to_strings();
        assert_eq!(
            strings,
            vec![
                "HDLJMKOBFPECGAIN",
                "AKOENCFDGILHPMBJ",
                "NGFBAPLIOJKMHCDE",
                "MIPCHJEGDBANKFOL",
                "IOHKFBDJAECGLNMP",
                "JBNFCMPEHLOIDKAG",
                "LEMDOGANPKFBCHJI",
                "GCAPILHKMNDJOBEF",
                "ENIGPOCMKABDJLFH",
                "PFKOGEJAIMHLBDNC",
                "DHJMKIBLCFNOEPGA",
                "CABLDHNFJGPEMIKO",
                "OPCAENKHBDGFIJLM",
                "BLEHJDIPNOMAFGCK",
                "KJGNBFMOLCIPAEHD",
                "FMDILAGCEHJKNOPB",
            ]
        );

        // Solve puzzle
        let mut solver = AnySolver::new_from(&board);
        assert!(solver.solve_logic());
        assert_eq!(solver.board, solution);

        let mut testfile = test_files_path();
        testfile.push("25-puzzle");

        let result = load_ksudoku(testfile);
        assert!(result.is_ok());
        let result = result.unwrap();
        assert!(result.puzzle_type == "sudoku");
        assert!(result.specific_type == "Plain");

        let size = (result.order as f64).sqrt() as u8;
        let mut board = AnyBoard::new(size);
        let load = board.load_ksudoku(&result.puzzle);
        assert!(load.is_ok());
        let mut solution = AnyBoard::new(size);
        let solution_load = solution.load_ksudoku(&result.solution);
        assert!(solution_load.is_ok());

        // Ok, puzzle is loaded!
        // board.display();
        let strings: Vec<String> = board.to_strings();
        assert_eq!(
            strings,
            vec![
                " T DPF    Y H R    WSX L ",
                "L QF     J X C G     UB D",
                " CK S LNRP  G  UTQO H MA ",
                "JG   H S XELDUPM F B   RT",
                "Y N RQ UCDOK AISJL HP G E",
                "F  OA N UK VQI HY B DT  C",
                "  S  A C VUE GJQ N I  R  ",
                "  CPD JGMIA   OELSU VBK  ",
                "  G LE D BHT XMW K PI Y  ",
                " EXIBOWFLY     VAMTJUGQS ",
                "G  HV TMI       EWF BR  O",
                " A JFUK W  P D  M GLXE N ",
                "R LN G             O QI F",
                " S TCYP B  H O  X JNAK M ",
                "M  XK ALJ       UHQ GP  Y",
                " VTRYBSEFM     KWOICJNLG ",
                "  U XD J WCN RTA E QY P  ",
                "  HQN COVTJ   EBGDL WSF  ",
                "  F  X Y LWB SVJ R T  O  ",
                "E  CJ R AN GOF XH V MD  B",
                "S V OI ANHDC TGLQJ YF X P",
                "CX   L K RFUYBWO V A   DQ",
                " FR H DQOC  K  INBW T UY ",
                "T JL     G I P F     OC W",
                " B KMV    Q J H    GRI E "
            ]
        );

        let strings: Vec<String> = solution.to_strings();
        assert_eq!(
            strings,
            vec![
                "UTODPFIBGAYMHQRNKCEWSXJLV",
                "LHQFIMYTEJSXWCNGVPAROUBKD",
                "XCKESWLNRPVFGJBUTQODHYMAI",
                "JGAVWHOSKXELDUPMIFYBQCNRT",
                "YMNBRQVUCDOKTAISJLXHPWGFE",
                "FJMOARNPUKLVQISHYGBXDTEWC",
                "KWSYTAHCXVUEBGJQONDILFRPM",
                "QRCPDTJGMIAYNWOELSUFVBKHX",
                "VNGULEQDSBHTFXMWCKRPIAYOJ",
                "HEXIBOWFLYRDPKCVAMTJUGQSN",
                "GQPHVCTMISXJANLYEWFKBRDUO",
                "OABJFUKVWQIPCDYRMTGLXESNH",
                "RYLNUGXHDETWVMKPBASOCQIJF",
                "WSETCYPRBFGHUOQDXIJNAKVML",
                "MIDXKNALJOBRSEFCUHQVGPWTY",
                "AVTRYBSEFMPQXHDKWOICJNLGU",
                "BOUSXDGJHWCNLRTAFEMQYVPIK",
                "IPHQNKCOVTJAMYEBGDLUWSFXR",
                "DKFMGXUYQLWBISVJPRNTEHOCA",
                "ELWCJPRIANKGOFUXHYVSMDTQB",
                "SUVWOIEANHDCRTGLQJKYFMXBP",
                "CXIGELMKTRFUYBWOSVPANJHDQ",
                "PFRAHJDQOCMSKVXINBWETLUYG",
                "TDJLQSBXYGNIEPAFRUHMKOCVW",
                "NBYKMVFWPUQOJLHTDXCGRIAES",
            ]
        );

        // solve puzzle
        let mut solver = AnySolver::new_from(&board);
        assert!(solver.solve_logic());
        assert_eq!(solver.board, solution);
    }
}
